from datetime import date

from django import forms
from django.core.exceptions import ValidationError
from django.forms.models import inlineformset_factory
from django.utils.translation import gettext_lazy as _, ngettext_lazy as ngettext

from ..models.journals import Journal, JournalEntry, JournalLeaderEntry, JournalTime
from ..models.roles import Leader
from ..models.schoolyear import SchoolYearDivision
from ..models.subjects import SubjectRegistrationParticipant, SubjectType
from ..models.timesheets import Timesheet, TimesheetPeriod
from ..utils import comma_separated, first_upper
from .fields import ReadonlyField
from .form import FormMixin
from .widgets import CheckboxSelectMultipleBootstrap


class JournalAdminForm(forms.ModelForm):
    class Meta:
        model = Journal
        exclude = ["subject", "school_year_division"]

    def __init__(self, data=None, *args, **kwargs):
        super().__init__(data, *args, **kwargs)

        # limit choices of leaders
        leaders_choices = self.fields["leaders"].widget.choices
        leaders_choices.queryset = leaders_choices.queryset.filter(subjects=self.subject)

        # limit choices of participants
        participants_qs = SubjectRegistrationParticipant.objects.filter(registration__subject=self.subject)

        if self.subject.subject_type.subject_type == SubjectType.COURSE:
            participants_qs = participants_qs.filter(
                registration__subject_variant__school_year_division=self.instance.school_year_division
            )

        participant_ids = set(participants_qs.values_list("id", flat=True))
        participant_ids |= set(self.instance.participants.values_list("id", flat=True))
        participant_ids |= set(
            SubjectRegistrationParticipant.objects.filter(
                journal_entries__journal=self.instance,
            ).values_list("id", flat=True),
        )
        self.fields["participants"].widget.choices.queryset = SubjectRegistrationParticipant.objects.filter(
            id__in=participant_ids
        )


class JournalCreateForm(FormMixin, forms.ModelForm):
    class Meta:
        model = Journal
        fields = ["name", "school_year_division"]

    def __init__(self, *args, **kwargs):
        self.subject = kwargs.pop("subject", None) or kwargs["instance"].subject
        super().__init__(*args, **kwargs)
        self.instance.subject = self.subject
        self.readonly_fields = [
            ReadonlyField(label=first_upper(self.subject.subject_type.name), value=self.subject),
        ]

        school_year_divisions = SchoolYearDivision.objects.filter(variants__subject=self.subject).distinct()
        if school_year_divisions.exists():
            self.fields["school_year_division"].widget.choices.queryset = school_year_divisions
        else:
            del self.fields["school_year_division"]


class JournalUpdateForm(FormMixin, forms.ModelForm):
    JournalTimeFormset = inlineformset_factory(
        Journal, JournalTime, fields=["day_of_week", "start", "end"], can_delete=True, extra=1
    )

    class Meta:
        model = Journal
        exclude = ["subject", "school_year_division"]
        widgets = {
            "leaders": CheckboxSelectMultipleBootstrap(),
            "participants": CheckboxSelectMultipleBootstrap(),
        }

    def __init__(self, *args, **kwargs):
        self.subject = kwargs.pop("subject", None) or kwargs["instance"].subject
        super().__init__(*args, **kwargs)
        self.instance.subject = self.subject
        self.readonly_fields = [
            ReadonlyField(label=first_upper(self.subject.subject_type.name), value=self.subject),
        ]

        participants_qs = SubjectRegistrationParticipant.objects.filter(registration__subject_id=self.subject.id)

        if self.instance.school_year_division:
            self.readonly_fields.append(
                ReadonlyField(label=first_upper(_("school year division")), value=self.instance.school_year_division),
            )
            participants_qs = participants_qs.filter(
                registration__subject_variant__school_year_division=self.instance.school_year_division
            )

        participant_ids = set(participants_qs.values_list("id", flat=True))
        participant_ids |= set(self.instance.participants.values_list("id", flat=True))
        participant_ids |= set(
            SubjectRegistrationParticipant.objects.filter(
                journal_entries__journal=self.instance,
            ).values_list("id", flat=True),
        )
        self.fields["participants"].widget.choices = tuple(
            (participant.id, participant)
            for participant in SubjectRegistrationParticipant.objects.filter(id__in=participant_ids)
        )

        self.fields["leaders"].widget.choices = tuple((leader.id, leader) for leader in self.subject.all_leaders)

        self.inline_formsets = [self.JournalTimeFormset(data=kwargs.get("data"), instance=kwargs.get("instance"))]

    def is_valid(self) -> bool:
        return all((super().is_valid(), *(inline_formset.is_valid() for inline_formset in self.inline_formsets)))

    def save(self, commit=True):
        obj = super().save(commit=commit)
        for inline_formset in self.inline_formsets:
            inline_formset.instance = obj
            inline_formset.save(commit=commit)
        return obj


class JournalLeaderEntryAdminForm(forms.ModelForm):
    class Meta:
        model = JournalLeaderEntry
        fields = ["start", "end"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.instance.timesheet.submitted:
            try:
                del self.fields["start"]
                del self.fields["end"]
            except KeyError:
                pass

    def clean(self):
        cleaned_data = super().clean()

        # readonly start end
        if self.instance.id and self.instance.timesheet.submitted:
            cleaned_data["start"] = self.instance.start
            cleaned_data["end"] = self.instance.end

        # check entry start and end
        errors = {}
        journal_entry = self.instance.journal_entry
        if cleaned_data["start"] < journal_entry.start:
            errors["start"] = _("The journal entry starts at {start}").format(
                start=journal_entry.start,
            )
        if cleaned_data["end"] > journal_entry.end:
            errors["end"] = _("The journal entry ends at {end}").format(
                end=journal_entry.end,
            )
        if errors:
            raise ValidationError(errors)

        return cleaned_data


class JournalLeaderEntryForm(FormMixin, JournalLeaderEntryAdminForm):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.readonly_fields = [
            ReadonlyField(label=_("Journal"), value=self.instance.journal_entry.journal),
            ReadonlyField(label=_("Date"), value=self.instance.journal_entry.date),
            ReadonlyField(label=_("Leader"), value=self.instance.timesheet.leader),
        ]
        if self.instance.timesheet.submitted:
            self.readonly_fields += [
                ReadonlyField(label=_("Start"), value=self.instance.start),
                ReadonlyField(label=_("End"), value=self.instance.end),
            ]


class JournalEntryAdminForm(forms.ModelForm):
    class Meta:
        model = JournalEntry
        fields = ["date", "start", "end", "agenda", "participants", "participants_instructed"]

    def __init__(self, *args, **kwargs):
        self.journal = kwargs.pop("journal", None) or kwargs["instance"].journal
        super().__init__(*args, **kwargs)
        self.instance.journal = self.journal

        if self.instance.id:
            d = self.instance.date
            self.readonly_fields = [
                ReadonlyField(label=_("Date"), value=d),
            ]
            try:
                del self.fields["date"]
            except KeyError:
                pass
        else:
            last = self.instance.journal.journal_entries.last()
            if last:
                last_end = last.datetime_end or last.date
            else:
                last_end = None
            next_time = self.instance.journal.get_next_time(last_end)
            if next_time:
                self.initial["date"] = next_time.date
                self.initial["start"] = next_time.start
                self.initial["end"] = next_time.end
            else:
                self.initial["date"] = date.today()
            try:
                d = self.fields["date"].clean(kwargs["data"]["date"])
            except (KeyError, TypeError, ValidationError):
                d = self.initial["date"]

        self.fields["participants"].widget.choices.queryset = self.instance.journal.get_valid_participants(d)
        self.fields["participants"].help_text = None

        self.fields["participants_instructed"].widget.choices.queryset = self.instance.journal.get_valid_participants(d)
        self.fields["participants_instructed"].help_text = None

    def clean(self):
        date = self.cleaned_data.get("date", self.instance.date)
        start = self.cleaned_data.get("start")
        end = self.cleaned_data.get("end")

        # start and end must be both set or both None
        if start is None and end is not None:
            raise ValidationError({"start": _("If You fill in start time, You must fill end time too.")})
        elif end is None and start is not None:
            raise ValidationError({"end": _("If You fill in end time, You must fill start time too.")})

        # check overlaping entries
        if start:
            qs = JournalEntry.objects.filter(
                journal=self.instance.journal,
                date=self.cleaned_data.get("date", self.instance.date),
                start__lt=end,
                end__gt=start,
            )
            if self.instance.id:
                qs = qs.exclude(id=self.instance.id)
            if qs.exists():
                raise ValidationError(_("An overlaping entry has already been added in the journal."))

        # check instructed participants
        participants = set(self.cleaned_data.get("participants", []))
        participants_instructed = set(self.cleaned_data.get("participants_instructed", []))
        not_present_participants_instructed = participants_instructed - participants
        if not_present_participants_instructed:
            self.add_error(
                "participants_instructed",
                _("Following participants were selected as instructed, but not selected as present: {}").format(
                    comma_separated(map(str, not_present_participants_instructed))
                ),
            )
        not_instructed_participants = []
        for participant in participants - participants_instructed:
            if not participant.instructed.filter(date__lt=date).exists():
                not_instructed_participants.append(participant)
        if not_instructed_participants:
            self.add_error(
                "participants",
                _("Following participants have not yet been instructed: {}").format(
                    comma_separated(map(str, not_instructed_participants))
                ),
            )

        # check submitted leader entries
        submitted_leader_entries = [e for e in self.instance.all_leader_entries if e.timesheet.submitted]
        if submitted_leader_entries:
            max_start = min(e.start for e in submitted_leader_entries)
            min_end = max(e.end for e in submitted_leader_entries)
            errors = {}
            if (start is None) or (start > max_start):
                errors["start"] = _("Some submitted timesheet entries start at {start}").format(
                    start=max_start,
                )
            if (end is None) or (end < min_end):
                errors["end"] = _("Some submitted timesheet entries end at {end}").format(
                    end=min_end,
                )
            if errors:
                raise ValidationError(errors)

        return self.cleaned_data


class JournalEntryForm(FormMixin, JournalEntryAdminForm):
    leaders = forms.ModelMultipleChoiceField(Leader.objects, label=_("Leaders"), required=False)
    alternates = forms.ModelMultipleChoiceField(Leader.objects, label=_("Alternates"), required=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.readonly_fields = [
            ReadonlyField(label=_("Journal"), value=self.journal),
        ]

        # only allow to select leaders or alterates with not submitted timesheets
        leaders = self.instance.journal.all_leaders
        alternates = [leader for leader in Leader.objects.all() if leader not in leaders]

        self.fields["leaders"].widget.choices = tuple((leader.id, leader) for leader in leaders)
        self.fields["leaders"].help_text = None
        self.fields["alternates"].widget.choices = tuple((leader.id, leader) for leader in alternates)
        self.fields["alternates"].help_text = None

        if self.instance.id:
            self.initial["leaders"] = [leader.id for leader in self.instance.all_leaders]
            self.initial["alternates"] = [leader.id for leader in self.instance.all_alternates]
        else:
            self.initial["leaders"] = [leader.id for leader in leaders]

    def clean(self):
        self.cleaned_data = super().clean()
        self.cleaned_entries = []
        self.deleted_entries = []

        d = self.cleaned_data.get("date", self.instance.date)
        if d is None:
            # no other validation makes sense without date
            return self.cleaned_data

        if (
            "start" in self.cleaned_data
            and "end" in self.cleaned_data
            and self.cleaned_data["start"]
            and self.cleaned_data["end"]
        ):
            errors = {}
            entries_by_leader = {
                "leaders": dict(
                    (entry.timesheet.leader, entry)
                    for entry in self.instance.all_leader_entries
                    if entry.timesheet.leader in self.instance.all_leaders
                ),
                "alternates": dict(
                    (entry.timesheet.leader, entry)
                    for entry in self.instance.all_leader_entries
                    if entry.timesheet.leader in self.instance.all_alternates
                ),
            }
            leaders_with_submitted_timesheets = {}
            period = TimesheetPeriod.objects.for_date(d)

            for group in ("leaders", "alternates"):
                leaders_with_submitted_timesheets[group] = []
                for leader in self.cleaned_data[group]:
                    if leader not in entries_by_leader[group]:
                        # try to create new leader entry
                        timesheet = Timesheet.objects.for_leader_and_date(leader=leader, date=d)
                        if timesheet.submitted:
                            # can not create entry
                            leaders_with_submitted_timesheets[group].append(leader)
                            continue
                        entry = JournalLeaderEntry()
                        entry.journal_entry = self.instance
                        entry.timesheet = timesheet
                        entry.start = self.cleaned_data["start"]
                        entry.end = self.cleaned_data["end"]
                    else:
                        # try to update existing leader entry
                        entry = entries_by_leader[group].pop(leader)
                        if not entry.timesheet.submitted:
                            if self.cleaned_data["start"] != self.instance.start:
                                if entry.start == self.instance.start:
                                    entry.start = self.cleaned_data["start"]
                                else:
                                    entry.start = max(entry.start, self.cleaned_data["start"])
                            if self.cleaned_data["end"] != self.instance.end:
                                if entry.end == self.instance.end:
                                    entry.end = self.cleaned_data["end"]
                                else:
                                    entry.end = min(entry.end, self.cleaned_data["end"])
                    # store cleaned entry, or delete it, if it is broken by the update
                    if entry.start < entry.end:
                        self.cleaned_entries.append(entry)
                    elif entry.id:
                        self.deleted_entries.append(entry)
                # try to delete stale entries
                for entry in entries_by_leader[group].values():
                    if entry.timesheet.submitted:
                        # can not delete entry
                        leaders_with_submitted_timesheets[group].append(entry.timesheet.leader)
                        continue
                    # store deleted entry
                    self.deleted_entries.append(entry)

            if leaders_with_submitted_timesheets["leaders"]:
                errors["leaders"] = ngettext(
                    "Leader {leaders} has already submitted timesheet for {period}.",
                    "Leaders {leaders} have already submitted timesheet for {period}.",
                    len(leaders_with_submitted_timesheets["leaders"]),
                ).format(
                    leaders=comma_separated(leaders_with_submitted_timesheets["leaders"]),
                    period=period.name,
                )
            if leaders_with_submitted_timesheets["alternates"]:
                errors["alternates"] = ngettext(
                    "Alternate {leaders} has already submitted timesheet for {period}.",
                    "Alternates {leaders} have already submitted timesheet for {period}.",
                    len(leaders_with_submitted_timesheets["alternates"]),
                ).format(
                    leaders=comma_separated(leaders_with_submitted_timesheets["alternates"]),
                    period=period.name,
                )
            if errors:
                raise ValidationError(errors)

        return self.cleaned_data

    def save(self, commit=True):
        self.instance = super().save(commit)
        for entry in self.cleaned_entries:
            entry.journal_entry = self.instance
            entry.save()
        for entry in self.deleted_entries:
            entry.delete()



class FormMixin(object):
    required_css_class = "required"
    use_get = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for f in self.fields:
            # add form-control to classes
            classes = set(self.fields[f].widget.attrs.get("class", "").split())
            classes.add("form-control")
            self.fields[f].widget.attrs["class"] = " ".join(classes)
            # add data-type
            if "data-type" not in self.fields[f].widget.attrs:
                self.fields[f].widget.attrs["data-type"] = self.fields[f].__class__.__name__



from django.contrib import admin

from ..models.organizations import Organization


@admin.register(Organization)
class OrganizationAdmin(admin.ModelAdmin):
    list_display = ("name", "address", "phone", "email", "company_num", "vat_number", "bank_account")



import uuid

from django.core.mail import EmailMultiAlternatives
from django.db import models
from django.template.loader import get_template
from django.urls import reverse_lazy as reverse
from django.utils.functional import cached_property
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from djangocms_text_ckeditor.fields import HTMLField
from filer.fields.file import FilerFileField

from ..conf import settings
from .leprikonsite import LeprikonSite


class Message(models.Model):
    created = models.DateTimeField(_("created"), editable=False, auto_now_add=True)
    sender = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        limit_choices_to={"is_staff": True},
        null=True,
        on_delete=models.SET_NULL,
        related_name="+",
        verbose_name=_("sender"),
    )
    subject = models.CharField(_("subject"), max_length=150)
    text = HTMLField(_("text"), default="")

    class Meta:
        app_label = "leprikon"
        ordering = ("-created",)
        verbose_name = _("message")
        verbose_name_plural = _("messages")

    def __str__(self):
        return "{}".format(self.subject)

    @cached_property
    def all_attachments(self):
        return list(self.attachments.all())

    @cached_property
    def all_recipients(self):
        return list(self.recipients.all())


class MessageRecipient(models.Model):
    slug = models.SlugField(editable=False)
    message = models.ForeignKey(
        Message, editable=False, on_delete=models.CASCADE, related_name="recipients", verbose_name=_("message")
    )
    recipient = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        on_delete=models.CASCADE,
        related_name="leprikon_messages",
        verbose_name=_("recipient"),
    )
    sent = models.DateTimeField(_("sent"), editable=False, auto_now_add=True)
    viewed = models.DateTimeField(_("viewed on site"), editable=False, null=True, default=None)
    sent_mail = models.DateTimeField(_("sent by email"), editable=False, null=True, default=None)

    class Meta:
        app_label = "leprikon"
        verbose_name = _("recipient")
        verbose_name_plural = _("recipients")
        ordering = ("-sent",)
        unique_together = (("message", "recipient"),)

    def __str__(self):
        return "{} ({})".format(self.recipient.get_username(), self.recipient.get_full_name())

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = uuid.uuid4()
            while MessageRecipient.objects.filter(slug=self.slug).exists():
                self.slug = uuid.uuid4()
        super().save(*args, **kwargs)

    def get_absolute_url(self):
        return reverse("leprikon:message_detail", args=(self.slug,))

    def send_mail(self):
        context = {
            "message_recipient": self,
            "site": LeprikonSite.objects.get_current(),
        }
        if self.message.sender_id:
            from_email = f'"{self.message.sender.get_full_name()}" <{settings.SERVER_EMAIL_PLAIN}>'
            reply_to = f'"{self.message.sender.get_full_name()}" <{self.message.sender.email}>'
        else:
            from_email = settings.SERVER_EMAIL
            reply_to = None
        msg = EmailMultiAlternatives(
            subject=self.message.subject,
            body=get_template("leprikon/message_mail.txt").render(context),
            from_email=from_email,
            to=[self.recipient.email],
            headers={"X-Mailer": "Leprikon (http://leprikon.cz/)"},
            reply_to=[reply_to],
        )
        msg.attach_alternative(get_template("leprikon/message_mail.html").render(context), "text/html")
        for attachment in self.message.attachments.all():
            msg.attach_file(attachment.file.file.path)
        msg.send()
        self.sent_mail = now()
        self.save()


class MessageAttachment(models.Model):
    message = models.ForeignKey(
        Message, on_delete=models.CASCADE, related_name="attachments", verbose_name=_("message")
    )
    file = FilerFileField(on_delete=models.CASCADE, related_name="+")

    class Meta:
        app_label = "leprikon"
        verbose_name = _("attachment")
        verbose_name_plural = _("attachments")

    def __str__(self):
        return str(self.file)



from django.contrib import admin
from django.utils.translation import gettext_lazy as _

from ..models.journals import Journal
from ..models.roles import Leader
from ..models.schoolyear import SchoolYear
from ..models.subjects import Subject, SubjectGroup, SubjectType


class SchoolYearListFilter(admin.FieldListFilter):
    def __init__(self, field, request, params, model, model_admin, field_path):
        try:
            request.school_year = SchoolYear.objects.get(year=params["year"])
        except (SchoolYear.DoesNotExist, ValueError, KeyError):
            pass
        self.school_year = request.school_year
        super().__init__(field, request, params, model, model_admin, field_path)

    def expected_parameters(self):
        return ["year"]

    def choices(self, cl):
        return [
            {
                "selected": school_year == self.school_year,
                "query_string": cl.get_query_string({"year": school_year.year}),
                "display": school_year,
            }
            for school_year in SchoolYear.objects.all()
        ]

    def queryset(self, request, queryset):
        return queryset.filter(**{self.field_path: self.school_year})


class ActiveListFilter(admin.SimpleListFilter):
    title = _("status")
    parameter_name = "status"

    def lookups(self, request, model_admin):
        return (
            (None, _("active")),
            ("inactive", _("inactive")),
        )

    def queryset(self, request, queryset):
        if self.value() == "inactive":
            return queryset.filter(active=False)
        else:
            return queryset.filter(active=True)

    def choices(self, cl):
        value = self.value()
        return [
            {
                "selected": value != "inactive",
                "query_string": cl.get_query_string({}, [self.parameter_name]),
                "display": _("active"),
            },
            {
                "selected": value == "inactive",
                "query_string": cl.get_query_string({self.parameter_name: "inactive"}),
                "display": _("inactive"),
            },
        ]


class ApprovedListFilter(admin.SimpleListFilter):
    title = _("approval")
    parameter_name = "approved"

    def lookups(self, request, model_admin):
        return (
            ("yes", _("approved")),
            ("no", _("unapproved")),
        )

    def queryset(self, request, queryset):
        if self.value() == "yes":
            return queryset.filter(approved__isnull=False)
        if self.value() == "no":
            return queryset.filter(approved__isnull=True)


class CanceledListFilter(admin.SimpleListFilter):
    title = _("cancelation")
    parameter_name = "canceled"

    def lookups(self, request, model_admin):
        return (
            (None, _("not canceled")),
            ("yes", _("canceled")),
        )

    def queryset(self, request, queryset):
        if self.value() == "yes":
            return queryset.filter(canceled__isnull=False)
        else:
            return queryset.filter(canceled__isnull=True)

    def choices(self, cl):
        value = self.value()
        return [
            {
                "selected": value != "yes",
                "query_string": cl.get_query_string({}, [self.parameter_name]),
                "display": _("active"),
            },
            {
                "selected": value == "yes",
                "query_string": cl.get_query_string({self.parameter_name: "yes"}),
                "display": _("canceled"),
            },
        ]


class SubjectTypeListFilter(admin.RelatedFieldListFilter):
    def __init__(self, field, request, params, model, model_admin, field_path):
        self.subject_types = SubjectType.objects.all()
        subject_type_type = getattr(model_admin, "subject_type_type", None)
        if subject_type_type:
            self.subject_types = self.subject_types.filter(subject_type=subject_type_type)
        super().__init__(field, request, params, model, model_admin, field_path)
        if subject_type_type:
            self.title = SubjectType.subject_type_type_labels[subject_type_type]
        request.leprikon_subject_type_id = self.lookup_val

    def field_choices(self, field, request, model_admin):
        return [(t.id, t.plural) for t in self.subject_types.all()]


class SubjectGroupListFilter(admin.RelatedFieldListFilter):
    def __init__(self, field, request, params, model, model_admin, field_path):
        subject_type_type = getattr(model_admin, "subject_type_type", None)
        if hasattr(request, "leprikon_subject_type_id") and request.leprikon_subject_type_id:
            self.groups = SubjectGroup.objects.filter(subject_types__id=request.leprikon_subject_type_id)
        elif subject_type_type:
            self.groups = SubjectGroup.objects.filter(subject_types__subject_type=subject_type_type)
        else:
            self.groups = SubjectGroup.objects.all()
        super().__init__(field, request, params, model, model_admin, field_path)
        request.leprikon_subject_group_id = self.lookup_val

    def field_choices(self, field, request, model_admin):
        return [(g.id, g.plural) for g in self.groups.distinct()]


class SubjectListFilter(admin.RelatedFieldListFilter):
    model = Subject

    def __init__(self, field, request, params, model, model_admin, field_path):
        self.subjects = self.model.objects.filter(school_year=request.school_year)
        subject_type_type = getattr(model_admin, "subject_type_type", None)
        if hasattr(request, "leprikon_subject_type_id") and request.leprikon_subject_type_id:
            self.subjects = self.subjects.filter(subject_type__id=request.leprikon_subject_type_id)
        elif subject_type_type:
            self.subjects = self.subjects.filter(subject_type__subject_type=subject_type_type)
        if hasattr(request, "leprikon_subject_group_id") and request.leprikon_subject_group_id:
            self.subjects = self.subjects.filter(subject_group__id=request.leprikon_subject_group_id)
        super().__init__(field, request, params, model, model_admin, field_path)
        if subject_type_type:
            self.title = SubjectType.subject_type_labels[subject_type_type]

    def field_choices(self, field, request, model_admin):
        return [(s.id, s.name) for s in self.subjects]


class JournalListFilter(admin.RelatedFieldListFilter):
    model = Journal

    def __init__(self, field, request, params, model, model_admin, field_path):
        self.journals = (
            self.model.objects.filter(subject__school_year=request.school_year)
            .select_related("subject")
            .order_by("subject__name", "name")
        )
        if hasattr(request, "leprikon_subject_id") and request.leprikon_subject_id:
            self.journals = self.journals.filter(subject_id=request.leprikon_subject_id)
        elif hasattr(request, "leprikon_subject_type_id") and request.leprikon_subject_type_id:
            self.journals = self.journals.filter(subject__subject_type__id=request.leprikon_subject_type_id)
        super().__init__(field, request, params, model, model_admin, field_path)

    def field_choices(self, field, request, model_admin):
        return [(journal.id, str(journal)) for journal in self.journals]


class LeaderListFilter(admin.RelatedFieldListFilter):
    def __init__(self, field, request, params, model, model_admin, field_path):
        self.leaders = Leader.objects.filter(school_years=request.school_year)
        super().__init__(field, request, params, model, model_admin, field_path)

    def field_choices(self, field, request, model_admin):
        return [(leader.id, leader) for leader in self.leaders]


class IsNullFieldListFilter(admin.FieldListFilter):
    def __init__(self, field, request, params, model, model_admin, field_path):
        self.lookup_kwarg = "%s__isnull" % field_path
        self.lookup_val = request.GET.get(self.lookup_kwarg)
        if not hasattr(field, "verbose_name") and hasattr(field, "related_model"):
            field.verbose_name = field.related_model._meta.verbose_name
        super().__init__(field, request, params, model, model_admin, field_path)
        if (
            self.used_parameters
            and self.lookup_kwarg in self.used_parameters
            and self.used_parameters[self.lookup_kwarg] in ("1", "0")
        ):
            self.used_parameters[self.lookup_kwarg] = bool(int(self.used_parameters[self.lookup_kwarg]))

    def expected_parameters(self):
        return [self.lookup_kwarg]

    def choices(self, changelist):
        for lookup, title in ((None, _("All")), ("0", _("Set")), ("1", _("Not set"))):
            yield {
                "selected": self.lookup_val == lookup,
                "query_string": changelist.get_query_string(
                    {
                        self.lookup_kwarg: lookup,
                    }
                ),
                "display": title,
            }



from django.db import models
from django.utils.translation import gettext_lazy as _

from ..conf import settings


class UserAgreement(models.Model):
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        primary_key=True,
        verbose_name=_("user"),
        related_name="agreement",
        on_delete=models.CASCADE,
    )
    granted = models.DateTimeField(_("time of agreement"), editable=False, auto_now=True)



from collections import namedtuple
from functools import lru_cache

from django.db.models import Sum
from django.template.response import TemplateResponse
from django.urls import reverse_lazy as reverse
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from ...forms.reports.events import EventPaymentsForm, EventPaymentsStatusForm, EventStatsForm
from ...models.agegroup import AgeGroup
from ...models.citizenship import Citizenship
from ...models.events import Event, EventRegistration
from ...models.roles import Participant
from ...models.subjects import SubjectPayment, SubjectRegistrationParticipant, SubjectType
from ...views.generic import FormView


class ReportEventPaymentsView(FormView):
    form_class = EventPaymentsForm
    template_name = "leprikon/reports/event_payments.html"
    title = _("Event payments")
    submit_label = _("Show")
    back_url = reverse("leprikon:report_list")

    def form_valid(self, form):
        context = form.cleaned_data
        context["form"] = form
        context["received_payments"] = SubjectPayment.objects.filter(
            target_registration__subject__subject_type__subject_type=SubjectType.EVENT,
            accounted__gte=context["date_start"],
            accounted__lte=context["date_end"],
        )
        context["returned_payments"] = SubjectPayment.objects.filter(
            source_registration__subject__subject_type__subject_type=SubjectType.EVENT,
            accounted__gte=context["date_start"],
            accounted__lte=context["date_end"],
        )
        context["received_payments_sum"] = context["received_payments"].aggregate(sum=Sum("amount"))["sum"] or 0
        context["returned_payments_sum"] = context["returned_payments"].aggregate(sum=Sum("amount"))["sum"] or 0
        context["sum"] = context["received_payments_sum"] - context["returned_payments_sum"]
        return TemplateResponse(self.request, self.template_name, self.get_context_data(**context))


class ReportEventPaymentsStatusView(FormView):
    form_class = EventPaymentsStatusForm
    template_name = "leprikon/reports/event_payments_status.html"
    title = _("Event payments status")
    submit_label = _("Show")
    back_url = reverse("leprikon:report_list")

    EventPaymentsStatusSums = namedtuple("EventPaymentsStatusSums", ("registrations", "status"))

    def form_valid(self, form):
        context = form.cleaned_data
        context["form"] = form
        context["reports"] = [
            self.Report(event, context["date"]) for event in Event.objects.filter(school_year=self.request.school_year)
        ]
        context["sum"] = self.EventPaymentsStatusSums(
            registrations=sum(len(r.registration_statuses) for r in context["reports"]),
            status=sum(r.status for r in context["reports"]),
        )
        return TemplateResponse(self.request, self.template_name, self.get_context_data(**context))

    class Report:
        def __init__(self, event, d):
            self.event = event
            self.date = d

        RegPaymentStatus = namedtuple("RegPaymentStatus", ("registration", "status"))

        @cached_property
        def registration_statuses(self):
            return [
                registration_status
                for registration_status in (
                    self.RegPaymentStatus(
                        registration=registration,
                        status=registration.get_payment_status(self.date),
                    )
                    for registration in EventRegistration.objects.filter(
                        subject=self.event,
                        approved__date__lte=self.date,
                    )
                )
                if registration_status.status.receivable
            ]

        @cached_property
        def status(self):
            return sum(rs.status for rs in self.registration_statuses)


class ReportEventStatsView(FormView):
    form_class = EventStatsForm
    template_name = "leprikon/reports/event_stats.html"
    title = _("Event statistics")
    submit_label = _("Show")
    back_url = reverse("leprikon:report_list")

    ReportItem = namedtuple("ReportItem", ("age_group", "all", "boys", "girls", "citizenships"))

    @lru_cache
    def get_event_days(self, event_id):
        event = Event.objects.get(id=event_id)
        return (event.end_date - event.start_date).days + 1

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs["school_year"] = self.request.school_year
        return kwargs

    def form_valid(self, form):
        d = form.cleaned_data["date"]
        paid_only = form.cleaned_data["paid_only"]
        paid_later = form.cleaned_data["paid_later"]
        approved_later = form.cleaned_data["approved_later"]
        unique_participants = form.cleaned_data["unique_participants"]
        min_days = form.cleaned_data["min_days"]
        context = form.cleaned_data
        context["form"] = form

        if approved_later:
            # approved registrations created by the date
            participants = SubjectRegistrationParticipant.objects.filter(
                registration__created__date__lte=d,
                registration__approved__isnull=False,
            )
        else:
            # registrations approved by the date
            participants = SubjectRegistrationParticipant.objects.filter(
                registration__approved__date__lte=d,
            )
        participants = (
            participants.filter(
                registration__subject__in=form.cleaned_data["events"],
            )
            .exclude(registration__canceled__date__lte=d)
            .select_related("registration", "age_group")
        )
        if paid_only:
            paid_date = None if paid_later else d
            participants = [
                participant
                for participant in participants
                if participant.registration.eventregistration.get_payment_status(paid_date).balance >= 0
            ]
        else:
            participants = list(participants)

        if min_days:
            participants = [p for p in participants if self.get_event_days(p.registration.subject_id) >= min_days]

        context["events_count"] = len(set(participant.registration.subject_id for participant in participants))

        if unique_participants:
            participants = list({p.key: p for p in participants}.values())

        citizenships = list(Citizenship.objects.all())
        context["citizenships"] = citizenships

        context["participants_counts"] = self.ReportItem(
            age_group=None,
            all=len(participants),
            boys=len([p for p in participants if p.gender == Participant.MALE]),
            girls=len([p for p in participants if p.gender == Participant.FEMALE]),
            citizenships=[
                len([p for p in participants if p.citizenship_id == citizenship.id]) for citizenship in citizenships
            ],
        )
        context["participants_counts_by_age_groups"] = []
        for age_group in AgeGroup.objects.all():
            parts = [p for p in participants if p.age_group == age_group]
            context["participants_counts_by_age_groups"].append(
                self.ReportItem(
                    age_group=age_group,
                    all=len(parts),
                    boys=len([p for p in parts if p.gender == Participant.MALE]),
                    girls=len([p for p in parts if p.gender == Participant.FEMALE]),
                    citizenships=[
                        len([p for p in parts if p.citizenship_id == citizenship.id]) for citizenship in citizenships
                    ],
                )
            )

        return TemplateResponse(self.request, self.template_name, self.get_context_data(**context))



from django import forms
from django.contrib import admin
from django.contrib.admin.widgets import AdminRadioSelect
from django.shortcuts import render
from django.templatetags.static import static
from django.urls import reverse
from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from ..models.orderables import Orderable, OrderableDiscount, OrderableRegistration
from ..models.schoolyear import SchoolYear
from ..models.subjects import SubjectType
from ..utils import attributes, currency
from .pdf import PdfExportAdminMixin
from .subjects import SubjectBaseAdmin, SubjectDiscountBaseAdmin, SubjectRegistrationBaseAdmin


@admin.register(Orderable)
class OrderableAdmin(SubjectBaseAdmin):
    subject_type_type = SubjectType.ORDERABLE
    registration_model = OrderableRegistration
    list_display = (
        "id",
        "code",
        "name",
        "subject_type",
        "get_groups_list",
        "get_leaders_list",
        "get_times_list",
        "duration",
        "place",
        "get_registering_link",
        "get_registrations_link",
        "journals_link",
        "icon",
        "note",
    )
    list_export = (
        "id",
        "school_year",
        "code",
        "name",
        "department",
        "subject_type",
        "registration_type",
        "get_groups_list",
        "get_leaders_list",
        "get_age_groups_list",
        "get_target_groups_list",
        "get_times_list",
        "place",
        "public",
        "min_registrations_count",
        "max_registrations_count",
        "get_approved_registrations_count",
        "get_unapproved_registrations_count",
        "note",
    )
    actions = SubjectBaseAdmin.actions + (
        "publish",
        "unpublish",
        "copy_to_school_year",
    )

    @attributes(short_description=_("Publish selected orderable events"))
    def publish(self, request, queryset):
        Orderable.objects.filter(id__in=[reg["id"] for reg in queryset.values("id")]).update(public=True)
        self.message_user(request, _("Selected orderable events were published."))

    @attributes(short_description=_("Unpublish selected orderable events"))
    def unpublish(self, request, queryset):
        Orderable.objects.filter(id__in=[reg["id"] for reg in queryset.values("id")]).update(public=False)
        self.message_user(request, _("Selected orderable events were unpublished."))

    @attributes(short_description=_("Copy selected orderable events to another school year"))
    def copy_to_school_year(self, request, queryset):
        class SchoolYearForm(forms.Form):
            school_year = forms.ModelChoiceField(
                label=_("Target school year"),
                help_text=_("All selected orderable events will be copied to selected school year."),
                queryset=SchoolYear.objects.all(),
            )

        if request.POST.get("post", "no") == "yes":
            form = SchoolYearForm(request.POST)
            if form.is_valid():
                school_year = form.cleaned_data["school_year"]
                for orderable in queryset.all():
                    orderable.copy_to_school_year(school_year)
                self.message_user(
                    request,
                    _("Selected orderable events were copied to school year {}.").format(school_year),
                )
                return
        else:
            form = SchoolYearForm()

        adminform = admin.helpers.AdminForm(
            form,
            [(None, {"fields": list(form.base_fields)})],
            {},
            None,
            model_admin=self,
        )

        return render(
            request,
            "leprikon/admin/action_form.html",
            dict(
                title=_("Select target school year"),
                opts=self.model._meta,
                adminform=adminform,
                media=self.media + adminform.media,
                action="copy_to_school_year",
                action_checkbox_name=admin.helpers.ACTION_CHECKBOX_NAME,
                select_across=request.POST["select_across"],
                selected=request.POST.getlist(admin.helpers.ACTION_CHECKBOX_NAME),
            ),
        )


@admin.register(OrderableRegistration)
class OrderableRegistrationAdmin(PdfExportAdminMixin, SubjectRegistrationBaseAdmin):
    subject_type_type = SubjectType.ORDERABLE
    actions = SubjectRegistrationBaseAdmin.actions + PdfExportAdminMixin.actions + ("add_discounts",)
    list_display = (
        SubjectRegistrationBaseAdmin.list_display[:3] + ("event_date",) + SubjectRegistrationBaseAdmin.list_display[3:]
    )

    @attributes(short_description=_("Add discounts to selected registrations"))
    def add_discounts(self, request, queryset):
        ABSOLUTE = "A"
        RELATIVE = "R"

        class DiscountForm(forms.Form):
            discount_type = forms.ChoiceField(
                label=_("Discount type"),
                choices=((ABSOLUTE, _("absolute amount")), (RELATIVE, _("relative amount in percents"))),
                widget=AdminRadioSelect(),
            )
            amount = forms.DecimalField(label=_("Amount or number of percents"))
            explanation = OrderableDiscount._meta.get_field("explanation").formfield()

        if request.POST.get("post", "no") == "yes":
            form = DiscountForm(request.POST)
            if form.is_valid():
                OrderableDiscount.objects.bulk_create(
                    OrderableDiscount(
                        registration=registration,
                        amount=(
                            form.cleaned_data["amount"]
                            if form.cleaned_data["discount_type"] == ABSOLUTE
                            else (registration.price - sum(discount.amount for discount in registration.all_discounts))
                            * form.cleaned_data["amount"]
                            / 100
                        ),
                        explanation=form.cleaned_data["explanation"],
                    )
                    for registration in queryset.all()
                )
                self.message_user(request, _("The discounts have been created for selected registrations."))
                return
        else:
            form = DiscountForm()

        adminform = admin.helpers.AdminForm(
            form,
            [(None, {"fields": list(form.base_fields)})],
            {},
            None,
            model_admin=self,
        )

        return render(
            request,
            "leprikon/admin/action_form.html",
            dict(
                title=_("Add discounts"),
                opts=self.model._meta,
                adminform=adminform,
                media=self.media + adminform.media,
                action="add_discounts",
                action_checkbox_name=admin.helpers.ACTION_CHECKBOX_NAME,
                select_across=request.POST["select_across"],
                selected=request.POST.getlist(admin.helpers.ACTION_CHECKBOX_NAME),
            ),
        )

    @attributes(short_description=_("discounts"))
    def discounts(self, obj: OrderableRegistration):
        return mark_safe(
            format_html(
                '<a href="{href_list}"><b>{amount}</b></a>'
                ' &nbsp; <a class="popup-link" href="{href_add}" style="background-position: 0 0" title="{title_add}">'
                '<img src="{icon_add}" alt="+"/></a>',
                href_list=reverse("admin:leprikon_orderablediscount_changelist") + f"?registration={obj.id}",
                amount=currency(obj.payment_status.discount),
                href_add=reverse("admin:leprikon_orderablediscount_add") + f"?registration={obj.id}",
                icon_add=static("admin/img/icon-addlink.svg"),
                title_add=_("add discount"),
            )
        )

    @attributes(short_description=_("received payments"))
    def received_payments(self, obj: OrderableRegistration):
        return mark_safe(
            format_html(
                '<a style="color: {color}" href="{href_list}" title="{title}"><b>{amount}</b></a>'
                ' &nbsp; <a class="popup-link" href="{href_add}" style="background-position: 0 0" title="{title_add}">'
                '<img src="{icon_add}" alt="+"/></a>',
                color=obj.payment_status.color,
                href_list=reverse("admin:leprikon_subjectreceivedpayment_changelist")
                + "?target_registration={}".format(obj.id),
                title=obj.payment_status.title,
                amount=currency(obj.payment_status.received),
                href_add=reverse("admin:leprikon_subjectreceivedpayment_add")
                + "?target_registration={}".format(obj.id),
                icon_add=static("admin/img/icon-addlink.svg"),
                title_add=_("add payment"),
            )
        )

    @attributes(short_description=_("discounts"))
    def discounts_export(self, obj: OrderableRegistration):
        return currency(obj.payment_status.discount)

    @attributes(short_description=_("total price"))
    def total_price_export(self, obj: OrderableRegistration):
        return currency(obj.payment_status.receivable)

    @attributes(short_description=_("received payments"))
    def received_payments_export(self, obj: OrderableRegistration):
        return currency(obj.payment_status.received)


@admin.register(OrderableDiscount)
class OrderableDiscountAdmin(SubjectDiscountBaseAdmin):
    actions = SubjectDiscountBaseAdmin.actions
    list_display = ("accounted", "registration", "subject", "amount_html", "explanation")
    list_export = ("accounted", "registration", "subject", "amount", "explanation")



from cms.menu_bases import CMSAttachMenu
from django.db.models.signals import post_delete, post_save
from django.urls import NoReverseMatch, reverse
from django.utils.translation import gettext_lazy as _
from menus.base import Menu, Modifier, NavigationNode
from menus.menu_pool import menu_pool
from social_django.context_processors import backends

from .models.subjects import SubjectType
from .utils import current_url, first_upper, url_with_back


@menu_pool.register_menu
class LeprikonMenu(CMSAttachMenu):
    name = _("Leprikon")

    def get_nodes(self, request):
        """
        This method is used to build the menu tree.
        """
        try:
            nodes = []
            nodes.append(
                NavigationNode(
                    _("Summary"),
                    reverse("leprikon:summary"),
                    len(nodes),
                    attr={"visible_for_anonymous": False},
                )
            )
            nodes.append(
                NavigationNode(
                    _("Reports"),
                    reverse("leprikon:report_list"),
                    len(nodes),
                    attr={"require_staff": True},
                )
            )
            nodes.append(
                NavigationNode(
                    _("Leader"),
                    reverse("leprikon:leader_summary"),
                    len(nodes),
                    attr={"require_leader": True},
                )
            )
            for subject_type in SubjectType.objects.all():
                nodes.append(
                    NavigationNode(
                        first_upper(_("My {subject_type}").format(subject_type=subject_type.plural)),
                        reverse("leprikon:subject_list_mine", kwargs={"subject_type": subject_type.slug}),
                        len(nodes),
                        attr={"require_leader": subject_type.slug},
                    )
                )
            nodes.append(
                NavigationNode(
                    _("Alternating"),
                    reverse("leprikon:alternating"),
                    len(nodes),
                    attr={"require_leader": True},
                )
            )
            nodes.append(
                NavigationNode(
                    _("Timesheets"),
                    reverse("leprikon:timesheet_list"),
                    len(nodes),
                    attr={"require_leader": True},
                )
            )
            nodes.append(
                NavigationNode(
                    _("Messages"),
                    reverse("leprikon:message_list"),
                    len(nodes),
                    attr={"visible_for_anonymous": False},
                )
            )
            nodes.append(
                NavigationNode(
                    _("My registrations"),
                    reverse("leprikon:registration_list"),
                    len(nodes),
                    attr={"visible_for_anonymous": False},
                )
            )
            nodes.append(
                NavigationNode(
                    _("Payments"),
                    reverse("leprikon:payment_list"),
                    len(nodes),
                    attr={"visible_for_anonymous": False},
                )
            )
            nodes.append(
                NavigationNode(
                    _("Donations"),
                    reverse("leprikon:donation_list"),
                    len(nodes),
                    attr={"visible_for_anonymous": False},
                )
            )
            nodes.append(
                NavigationNode(
                    _("Participants"),
                    reverse("leprikon:participant_list"),
                    len(nodes),
                    attr={"visible_for_anonymous": False},
                )
            )
            nodes.append(
                NavigationNode(
                    _("Billing information"),
                    reverse("leprikon:billing_info_list"),
                    len(nodes),
                    attr={"visible_for_anonymous": False},
                )
            )
            return nodes
        except NoReverseMatch:
            return []


@menu_pool.register_menu
class UserMenu(Menu):
    def get_nodes(self, request):
        """
        This method is used to build the menu tree.
        """
        nodes = []

        # anonymous user
        nodes.append(
            NavigationNode(
                _("Log in"),
                reverse("leprikon:user_login"),
                len(nodes),
                attr={"visible_for_authenticated": False, "add_url_back": True},
            )
        )
        nodes.append(
            NavigationNode(
                _("Create account"),
                reverse("leprikon:user_create"),
                len(nodes),
                attr={"visible_for_authenticated": False, "add_url_back": True},
            )
        )
        nodes.append(
            NavigationNode(
                _("Reset password"),
                reverse("leprikon:password_reset"),
                len(nodes),
                attr={"visible_for_authenticated": False, "add_url_back": True},
            )
        )

        # user settings
        nodes.append(
            NavigationNode(
                _("Change user details"),
                reverse("leprikon:user_update"),
                len(nodes),
                attr={"visible_for_anonymous": False, "add_url_back": True},
            )
        )
        nodes.append(
            NavigationNode(
                _("Change e-mail"),
                reverse("leprikon:user_email"),
                len(nodes),
                attr={"visible_for_anonymous": False, "add_url_back": True},
            )
        )
        nodes.append(
            NavigationNode(
                _("Change password"),
                reverse("leprikon:user_password"),
                len(nodes),
                attr={"visible_for_anonymous": False, "add_url_back": True},
            )
        )
        nodes.append(
            NavigationNode(
                _("Link with Google account"),
                reverse("social:begin", args=("google-oauth2",)),
                len(nodes),
                attr={
                    "visible_for_anonymous": False,
                    "add_url_back": True,
                    "social_backend": "google-oauth2",
                    "social_btn": "google",
                },
            )
        )
        nodes.append(
            NavigationNode(
                _("Link with Facebook account"),
                reverse("social:begin", args=("facebook",)),
                len(nodes),
                attr={
                    "visible_for_anonymous": False,
                    "add_url_back": True,
                    "social_backend": "facebook",
                    "social_btn": "facebook",
                },
            )
        )
        nodes.append(
            NavigationNode(
                _("Log out"),
                reverse("leprikon:user_logout"),
                len(nodes),
                attr={"visible_for_anonymous": False},
            )
        )
        return nodes


@menu_pool.register_modifier
class LeprikonModifier(Modifier):
    def modify(self, request, nodes, namespace, root_id, post_cut, breadcrumb):
        if post_cut or breadcrumb:
            return nodes
        final = []
        backends_not_associated = backends(request)["backends"]["not_associated"]
        for node in nodes:
            show = True
            require_leader = node.attr.get("require_leader", False)
            if require_leader:
                kwargs = {"school_year": request.school_year}
                if isinstance(require_leader, str):
                    kwargs["subject_type__slug"] = require_leader
                if not request.leader or not request.leader.subjects.filter(**kwargs).exists():
                    show = False
            if node.attr.get("require_staff", False) and not request.user.is_staff:
                show = False
            if node.attr.get("social_backend") and (node.attr.get("social_backend") not in backends_not_associated):
                show = False
            if show:
                if node.attr.get("add_url_back", False):
                    node.url = url_with_back(node.url, current_url(request))
                final.append(node)
            else:
                if node.parent and node in node.parent.children:
                    node.parent.children.remove(node)
                continue
        return final


@menu_pool.register_modifier
class NamespaceModifier(Modifier):
    def modify(self, request, nodes, namespace, root_id, post_cut, breadcrumb):
        if namespace and not (post_cut or breadcrumb):
            namespaces = namespace.split(",")
            return [node for node in nodes if node.namespace.split(":")[0] in namespaces]
        else:
            return nodes


# clear menu cache with each reload
try:
    menu_pool.clear()
except Exception:
    # menu_pool.clear() uses database,
    # but this code may be executed
    # before the database is created
    pass


def invalidate_menu_cache(sender, **kwargs):
    menu_pool.clear()


post_save.connect(invalidate_menu_cache, sender=SubjectType)
post_delete.connect(invalidate_menu_cache, sender=SubjectType)



from collections import OrderedDict
from decimal import Decimal

from bankreader.models import Transaction as BankreaderTransaction
from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator
from django.db import models, transaction
from django.urls import reverse
from django.utils import formats, timezone
from django.utils.functional import cached_property
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _
from django_pays.models import Payment as PaysPayment

from ..conf import settings
from ..utils import attributes, currency
from .leprikonsite import LeprikonSite
from .organizations import Organization
from .pdfmail import PdfExportAndMailMixin


class TransactionManager(models.Manager):
    def get_queryset(self):
        qs = super().get_queryset()
        if self.model.transaction_types:
            qs = qs.filter(transaction_type__in=self.model.transaction_types)
        return qs


class AbstractTransaction(models.Model):
    accounted = models.DateTimeField(_("accounted time"), default=timezone.now)
    accounted_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        null=True,
        on_delete=models.PROTECT,
        related_name="+",
        verbose_name=_("accounted by"),
    )
    last_updated = models.DateTimeField(_("last updated time"), editable=False, null=True)
    last_updated_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        editable=False,
        null=True,
        on_delete=models.PROTECT,
        related_name="+",
        verbose_name=_("last updated by"),
    )

    class Meta:
        abstract = True

    def save(self, *args, **kwargs):
        self.clean()
        super().save(*args, **kwargs)

    def clean(self):
        errors = {}
        try:
            super().clean()
        except ValidationError as e:
            errors = e.update_error_dict(errors)

        if self.amount == 0:
            errors["amount"] = [_("Amount can not be zero.")]
        if self.accounted:
            max_closure_date = LeprikonSite.objects.get_current().max_closure_date
            if max_closure_date and self.accounted.date() <= max_closure_date:
                errors["accounted"] = [
                    _("Date must be after the last account closure ({}).").format(formats.date_format(max_closure_date))
                ]
        if errors:
            raise ValidationError(errors)


class Transaction(PdfExportAndMailMixin, AbstractTransaction):
    PAYMENT_CASH = "PAYMENT_CASH"
    PAYMENT_BANK = "PAYMENT_BANK"
    PAYMENT_ONLINE = "PAYMENT_ONLINE"
    RETURN_CASH = "RETURN_CASH"
    RETURN_BANK = "RETURN_BANK"
    TRANSFER = "TRANSFER"
    DONATION_CASH = "DONATION_CASH"
    DONATION_BANK = "DONATION_BANK"
    DONATION_ONLINE = "DONATION_ONLINE"
    DONATION_TRANSFER = "DONATION_TRANSFER"
    TRANSACTION_TYPE_LABELS = OrderedDict(
        [
            (PAYMENT_CASH, _("payment - cash")),
            (PAYMENT_BANK, _("payment - bank")),
            (PAYMENT_ONLINE, _("payment - online")),
            (RETURN_CASH, _("returned payment - cash")),
            (RETURN_BANK, _("returned payment - bank")),
            (TRANSFER, _("transfer between registrations")),
            (DONATION_CASH, _("donation - cash")),
            (DONATION_BANK, _("donation - bank")),
            (DONATION_ONLINE, _("donation - online")),
            (DONATION_TRANSFER, _("transfer from registration to donation")),
        ]
    )
    PAYMENTS = frozenset({PAYMENT_CASH, PAYMENT_BANK, PAYMENT_ONLINE, TRANSFER})
    RETURNS = frozenset({RETURN_CASH, RETURN_BANK, TRANSFER, DONATION_TRANSFER})
    SUBJECTS = frozenset(PAYMENTS.union(RETURNS))
    DONATIONS = frozenset({DONATION_CASH, DONATION_BANK, DONATION_ONLINE, DONATION_TRANSFER})
    BANKS = frozenset({PAYMENT_BANK, RETURN_BANK, DONATION_BANK})
    ONLINES = frozenset({PAYMENT_ONLINE, DONATION_ONLINE})

    # manager setup
    objects = TransactionManager()
    object_name = "transaction"
    transaction_types = None  # all

    # basic fields
    transaction_type = models.CharField(_("payment type"), max_length=30, choices=TRANSACTION_TYPE_LABELS.items())
    amount = models.DecimalField(
        _("amount"),
        decimal_places=2,
        max_digits=10,
        validators=[MinValueValidator(Decimal("0.01"))],
    )
    note = models.CharField(_("note"), max_length=300, blank=True, default="")

    # not editable
    mail_sent = models.DateTimeField(_("mail sent"), editable=False, null=True)

    # subject payment fields
    source_registration = models.ForeignKey(
        "SubjectRegistration",
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        related_name="returned_payments",
        verbose_name=_("source registration"),
        help_text=_("The payment will be deducted from this registration."),
    )
    target_registration = models.ForeignKey(
        "SubjectRegistration",
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        related_name="received_payments",
        verbose_name=_("target registration"),
        help_text=_("The payment will be added to this registration."),
    )

    # donation fields
    donor = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        related_name="leprikon_donations",
        verbose_name=_("donor user"),
    )
    organization = models.ForeignKey(
        Organization,
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        related_name="donations",
        verbose_name=_("organization"),
    )

    # bank fields
    bankreader_transaction = models.OneToOneField(
        BankreaderTransaction,
        blank=True,
        null=True,
        on_delete=models.PROTECT,
        verbose_name=_("bank account transaction"),
    )
    pays_payment = models.OneToOneField(
        PaysPayment,
        blank=True,
        editable=False,
        null=True,
        on_delete=models.PROTECT,
        verbose_name=_("online payment"),
    )

    class Meta:
        app_label = "leprikon"
        ordering = ("accounted",)
        verbose_name = _("transaction")
        verbose_name_plural = _("transactions")

    def clean(self):
        errors = {}
        try:
            super().clean()
        except ValidationError as e:
            errors = e.update_error_dict(errors)

        # prefill values
        if self.transaction_type == self.DONATION_TRANSFER and self.source_registration is not None:
            if not self.donor:
                self.donor = self.source_registration.user
            if not self.organization:
                self.organization = self.source_registration.organization

        mandatory_fields = []
        forbidden_fields = []

        if self.transaction_type in self.PAYMENTS:
            mandatory_fields.append("target_registration")
        elif self.transaction_type:
            forbidden_fields.append("target_registration")

        if self.transaction_type in self.RETURNS:
            mandatory_fields.append("source_registration")
        elif self.transaction_type:
            forbidden_fields.append("source_registration")

        if self.transaction_type in self.DONATIONS:
            mandatory_fields.append("donor")
            mandatory_fields.append("organization")
        elif self.transaction_type:
            forbidden_fields.append("donor")
            forbidden_fields.append("organization")

        if self.transaction_type and self.transaction_type not in self.BANKS:
            forbidden_fields.append("bankreader_transaction")

        if self.transaction_type and self.transaction_type not in self.ONLINES:
            forbidden_fields.append("pays_payment")

        for field in mandatory_fields:
            if not getattr(self, field):
                errors.setdefault(field, []).append(
                    _("This field is mandatory for transaction type {transaction_type}.").format(
                        transaction_type=self.transaction_type_label
                    )
                )
        for field in forbidden_fields:
            if getattr(self, field):
                errors.setdefault(field, []).append(
                    _("This field is forbidden for transaction type {transaction_type}.").format(
                        transaction_type=self.transaction_type_label
                    )
                )
        if errors:
            raise ValidationError(errors)

    def __str__(self):
        return f"{self.transaction_type_label} {currency(self.amount)}"

    @attributes(short_description=_("payment type"))
    @cached_property
    def transaction_type_label(self):
        return self.TRANSACTION_TYPE_LABELS.get(self.transaction_type, "-")

    def get_absolute_url(self):
        return reverse(f"leprikon:{self.object_name}_pdf", kwargs={"pk": self.pk, "slug": f"{self.slug}"})

    @cached_property
    def slug(self):
        return slugify(f"{self}-{self.id}")

    def send_mail(self, event="received"):
        with transaction.atomic():
            self.mail_sent = timezone.now()
            self.save()
            super().send_mail(event)



from io import BytesIO

from django.http import HttpResponse
from django.urls import path, reverse
from django.utils.safestring import mark_safe
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _
from pypdf import PdfReader, PdfWriter

from ..utils import attributes


class PdfExportAdminMixin:
    actions = ("export_pdf",)
    pdf_event = "pdf"

    @attributes(short_description=_("Export selected items in single PDF"))
    def export_pdf(self, request, queryset):
        # create PDF
        writer = PdfWriter()
        for obj in queryset.iterator():
            pdf_data = BytesIO()
            obj.write_pdf(self.pdf_event, pdf_data)
            pdf_data.seek(0)
            pdf = PdfReader(pdf_data)
            for i in range(len(pdf.pages)):
                writer.add_page(pdf.pages[i])

        # create PDF response object
        response = HttpResponse(content_type="application/pdf")
        response["Content-Disposition"] = 'attachment; filename="{}.pdf"'.format(
            slugify(self.model._meta.verbose_name_plural),
        )

        # write PDF to response
        writer.write(response)

        return response

    def get_urls(self):
        urls = super().get_urls()
        return [
            path(
                "<int:obj_id>-<event>.pdf",
                self.admin_site.admin_view(self.pdf),
                name="{}_{}_event_pdf".format(self.model._meta.app_label, self.model._meta.model_name),
            ),
            path(
                "<int:obj_id>.pdf",
                self.admin_site.admin_view(self.pdf),
                name="{}_{}_pdf".format(self.model._meta.app_label, self.model._meta.model_name),
            ),
        ] + urls

    def pdf(self, request, obj_id, event=pdf_event):
        obj = self.get_object(request, obj_id)

        # create PDF response object
        response = HttpResponse(content_type="application/pdf")
        response["Content-Disposition"] = 'filename="{}"'.format(obj.get_pdf_filename(event))

        # write PDF to response
        return obj.write_pdf(event, response)

    @attributes(short_description=_("download"))
    def download_tag(self, obj):
        return mark_safe(
            '<a href="{}">PDF</a>'.format(
                reverse(
                    "admin:{}_{}_pdf".format(self.model._meta.app_label, self.model._meta.model_name),
                    args=(obj.id,),
                )
            )
        )



# Generated by Django 3.2.16 on 2023-01-08 18:48

from django.db import migrations, models


def ordered_periods(periods):
    for order, period in enumerate(periods, 1):
        period.order = order
        yield period


def set_schoolyearperiod_order(apps, schema_editor):
    SchoolYearDivision = apps.get_model("leprikon", "SchoolYearDivision")
    SchoolYearPeriod = apps.get_model("leprikon", "SchoolYearPeriod")
    for school_year_division in SchoolYearDivision.objects.all():
        SchoolYearPeriod.objects.bulk_update(
            ordered_periods(school_year_division.periods.order_by("start", "id")), ["order"]
        )


class Migration(migrations.Migration):

    dependencies = [
        ("leprikon", "0073_add_limit_choices_to"),
    ]

    operations = [
        migrations.AddField(
            model_name="schoolyearperiod",
            name="order",
            field=models.IntegerField(blank=True, default=0, verbose_name="order"),
        ),
        migrations.RunPython(set_schoolyearperiod_order),
        migrations.AlterModelOptions(
            name="schoolyearperiod",
            options={
                "ordering": ("order",),
                "verbose_name": "school year period",
                "verbose_name_plural": "school year periods",
            },
        ),
    ]



from django import forms
from django.utils.translation import gettext_lazy as _

from ..conf import settings
from ..models.schoolyear import SchoolYear


class SchoolYearForm(forms.Form):
    school_year = forms.ModelChoiceField(
        label=_("School year"),
        queryset=SchoolYear.objects.all(),
        empty_label=None,
    )
    back = settings.LEPRIKON_PARAM_BACK

    def __init__(self, request, *args, **kwargs):
        if request:
            kwargs["initial"] = {"school_year": request.school_year}
        super().__init__(*args, **kwargs)
        if request and not request.user.is_staff:
            self.fields["school_year"].queryset = self.fields["school_year"].queryset.filter(active=True)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-03-21 12:57
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0048_birth_date'),
    ]

    operations = [
        migrations.AddField(
            model_name='courselistplugin',
            name='target_groups',
            field=models.ManyToManyField(blank=True, help_text='Keep empty to skip searching by target groups.', related_name='_courselistplugin_target_groups_+', to='leprikon.TargetGroup', verbose_name='target groups'),
        ),
        migrations.AddField(
            model_name='eventlistplugin',
            name='target_groups',
            field=models.ManyToManyField(blank=True, help_text='Keep empty to skip searching by target groups.', related_name='_eventlistplugin_target_groups_+', to='leprikon.TargetGroup', verbose_name='target groups'),
        ),
    ]



from cms.plugin_pool import plugin_pool
from django.utils.translation import gettext as _

from ..models.orderables import FilteredOrderableListPlugin, OrderableListPlugin, OrderablePlugin
from .base import LeprikonPluginBase


@plugin_pool.register_plugin
class LeprikonOrderablePlugin(LeprikonPluginBase):
    name = _("Orderable event")
    model = OrderablePlugin
    raw_id_fields = ("orderable",)

    def get_render_template(self, context, instance, placeholder):
        return "leprikon/cms/orderable/%s.html" % instance.template


@plugin_pool.register_plugin
class LeprikonOrderableListPlugin(LeprikonPluginBase):
    name = _("Orderable list")
    model = OrderableListPlugin
    filter_horizontal = ("age_groups", "target_groups", "groups", "leaders")

    def get_render_template(self, context, instance, placeholder):
        return "leprikon/cms/orderable_list/%s.html" % instance.template


@plugin_pool.register_plugin
class LeprikonFilteredOrderableListPlugin(LeprikonPluginBase):
    name = _("Orderable list with search form")
    model = FilteredOrderableListPlugin
    render_template = "leprikon/cms/orderable_list_filtered.html"



from cms.plugin_base import CMSPluginBase
from django.utils.translation import gettext as _


class LeprikonPluginBase(CMSPluginBase):
    cache = False
    module = _("Leprikon")
    text_enabled = True

    def render(self, context, instance, placeholder):
        context = super().render(context, instance, placeholder)
        if hasattr(instance, "render"):
            context.update(instance.render(context))
        return context



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-03-26 20:44
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0050_orderables'),
    ]

    operations = [
        migrations.AlterField(
            model_name='subjectregistrationgroup',
            name='name',
            field=models.CharField(blank=True, default='', max_length=150, verbose_name='group name'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2021-06-03 02:40
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import leprikon.models.fields


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("leprikon", "0060_transactions"),
    ]

    operations = [
        migrations.CreateModel(
            name="RefundRequest",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "requested",
                    models.DateTimeField(
                        default=django.utils.timezone.now, editable=False, verbose_name="requested time"
                    ),
                ),
                ("bank_account", leprikon.models.fields.BankAccountField(verbose_name="bank account number")),
                (
                    "registration",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="refund_request",
                        to="leprikon.SubjectRegistration",
                        verbose_name="registration",
                    ),
                ),
                (
                    "requested_by",
                    models.ForeignKey(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="+",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="requested by",
                    ),
                ),
            ],
            options={
                "verbose_name": "refund request",
                "verbose_name_plural": "refund requests",
                "ordering": ("requested",),
            },
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.24 on 2019-09-28 06:27
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('leprikon', '0034_subject_registration_group'),
    ]

    operations = [
        migrations.AlterField(
            model_name='course',
            name='school_year_division',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='courses', to='leprikon.SchoolYearDivision', verbose_name='school year division'),
        ),
        migrations.AlterField(
            model_name='coursediscount',
            name='registration',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='discounts', to='leprikon.CourseRegistration', verbose_name='registration'),
        ),
        migrations.AlterField(
            model_name='eventdiscount',
            name='registration',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='discounts', to='leprikon.EventRegistration', verbose_name='registration'),
        ),
        migrations.AlterField(
            model_name='subjectpayment',
            name='bankreader_transaction',
            field=models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, to='bankreader.Transaction', verbose_name='bank account transaction'),
        ),
        migrations.AlterField(
            model_name='subjectpayment',
            name='pays_payment',
            field=models.OneToOneField(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.PROTECT, to='django_pays.Payment', verbose_name='online payment'),
        ),
    ]



from django.db import migrations, models
import django.db.models.deletion
from leprikon.models.subjects import SubjectType


def set_journal_school_year_division(apps, schema_editor):
    SchoolYearDivision = apps.get_model("leprikon", "SchoolYearDivision")
    JournalEntry = apps.get_model("leprikon", "JournalEntry")
    Journal = apps.get_model("leprikon", "Journal")

    for journal in Journal.objects.filter(subject__subject_type__subject_type=SubjectType.COURSE):
        school_year_divisions = list(
            SchoolYearDivision.objects.filter(
                course_registrations__participants__in=journal.participants.all()
            ).distinct()
        )

        if len(school_year_divisions) == 0:
            journal.school_year_division = journal.subject.course.school_year_division
            journal.save()
            continue

        if len(school_year_divisions) == 1:
            journal.school_year_division = school_year_divisions[0]
            journal.save()
            continue

        # split journals by school_year_divisions
        for school_year_division in school_year_divisions:
            new_journal = Journal.objects.create(
                subject=journal.subject,
                name=f"{journal.name}, {school_year_division.name}"[:150],
                school_year_division=school_year_division,
                risks=journal.risks,
                plan=journal.plan,
                evaluation=journal.evaluation,
            )
            new_journal.leaders.set(journal.leaders.all())
            new_journal.participants.set(
                journal.participants.filter(registration__courseregistration__school_year_division=school_year_division)
            )
            for entry in journal.journal_entries.all():
                new_entry = JournalEntry.objects.create(
                    journal=new_journal,
                    date=entry.date,
                    start=entry.start,
                    end=entry.end,
                    agenda=entry.agenda,
                )
                new_entry.participants.set(
                    entry.participants.filter(
                        registration__courseregistration__school_year_division=school_year_division
                    )
                )
                new_entry.participants_instructed.set(
                    entry.participants_instructed.filter(
                        registration__courseregistration__school_year_division=school_year_division
                    )
                )
                entry.leader_entries.update(journal_entry=new_entry)
            journal.journal_entries.all().delete()
        journal.delete()


class Migration(migrations.Migration):
    dependencies = [
        ("leprikon", "0074_schoolyear_period_order"),
    ]

    operations = [
        migrations.AddField(
            model_name="journal",
            name="school_year_division",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="journals",
                to="leprikon.schoolyeardivision",
                verbose_name="school year division",
            ),
        ),
        migrations.RunPython(set_journal_school_year_division),
    ]
